@@ -1,10 +1,14 @@
-# Copyright 2014 SAP AG or an SAP affiliate company.
+# Copyright 2015 SAP AG or an SAP affiliate company.
 # 
 
 import operator
 import re
+import decimal
+import itertools
 
-from sqlalchemy.sql import compiler, expression, text, bindparam
+import sqlanydb
+
+from sqlalchemy.sql import compiler, expression, text, column, bindparam
 from sqlalchemy.engine import default, base, reflection, url
 from sqlalchemy import types as sqltypes
 from sqlalchemy.sql import operators as sql_operators
@@ -15,7 +19,7 @@ from sqlalchemy.types import CHAR, VARCHAR, TIME, NCHAR, NVARCHAR,\
                             TEXT, DATE, DATETIME, FLOAT, NUMERIC,\
                             BIGINT, INT, INTEGER, SMALLINT, BINARY,\
                             VARBINARY, DECIMAL, TIMESTAMP, Unicode,\
-                            UnicodeText, REAL
+                            UnicodeText, REAL, LargeBinary
 
 RESERVED_WORDS = set([
     "add", "all", "alter", "and",
@@ -76,6 +80,21 @@ RESERVED_WORDS = set([
     "within", "work", "writetext", "xml"
     ])
 
+class SQLAnyNoPrimaryKeyError(Exception):
+    """ exception that is raised when trying to load the primary keys for a 
+    table that does not have any columns marked as being a primary key. 
+    As noted in this documentation: 
+    http://docs.sqlalchemy.org/en/latest/faq.html#how-do-i-map-a-table-that-has-no-primary-key
+    if a table has fully duplicate rows, and has no primary key, it cannot be mapped.
+    Since we can't tell if a table has rows that are 'supposed' to act like a primary key,
+    we just throw an exception and hopes the user adds primary keys to the table instead.
+    """
+
+    def __init__(self, message, table_name):
+
+        super(SQLAnyNoPrimaryKeyError, self).__init__(message)
+
+        self.table_name = table_name
 
 class _SQLAnyUnitypeMixin(object):
     """these types appear to return a buffer object."""
@@ -204,6 +223,7 @@ ischema_names = {
     'varbinary': VARBINARY,
     'image': IMAGE,
     'bit': BIT,
+    "long binary": LargeBinary,
 
 # not in documentation for ASE 15.7
     'long varchar': TEXT,  # TODO
@@ -213,6 +233,19 @@ ischema_names = {
 }
 
 
+# converter function, only argument is the value returned
+# from the database that we want to convert
+_decimal_converter = lambda decAsString: decimal.Decimal(decAsString) if decAsString is not None else None   
+
+# list of types to have converted, and the callable that sqlanydb calls when
+# it needs to convert said type
+_converter_list = [(sqlanydb.DT_DECIMAL, _decimal_converter)]
+
+
+# register any converters we have with sqlanydb
+list(itertools.starmap(lambda x, y: sqlanydb.register_converter(x, y), _converter_list))
+
+
 class SQLAnyInspector(reflection.Inspector):
 
     def __init__(self, conn):
@@ -277,8 +310,8 @@ class SQLAnySQLCompiler(compiler.SQLCompiler):
         'milliseconds': 'millisecond'
     })
 
-    def get_select_precolumns(self, select):
-        s = "DISTINCT" if select._distinct else ""
+    def get_select_precolumns(self, select, **kw ):
+        s = "DISTINCT " if select._distinct else ""
         if select._limit:
             if select._limit == 1:
                 s += "FIRST "
@@ -289,12 +322,16 @@ class SQLAnySQLCompiler(compiler.SQLCompiler):
                 # SQL Anywhere doesn't allow "start at" without "top n"
                 s += "TOP ALL "
             s += "START AT %s " % (select._offset + 1,)
-        return s
+        if s != '':
+            return s
+        return compiler.SQLCompiler.get_select_precolumns(
+            self, select, **kw)
+
 
     def get_from_hint_text(self, table, text):
         return text
 
-    def limit_clause(self, select):
+    def limit_clause(self, select, **kw):
         # Limit in sybase is after the select keyword
         return ""
 
@@ -381,17 +418,27 @@ class SQLAnyDialect(default.DefaultDialect):
     postfetch_lastrowid = True
     supports_multivalues_insert = True
 
+    # if not present, then sqlalchemy expects a float when dealing with 'Numeric' decimal types
+    # but by default sqlanydb returns them as strings, which freaks sqlalchemy out
+    # so we return them as Decimal objects, by use of a converter function and 
+    # `sqlanydb.register_converter()`
+    supports_native_decimal = True 
+
+
+
     @classmethod
     def dbapi(self):
-        import sqlanydb
         return sqlanydb
 
     @property
     def driver(self):
-        import sqlanydb
         return sqlanydb
 
     def create_connect_args(self, url):
+
+        # get extra options
+        dialect_opts = dict(url.query)
+
         opts = url.translate_connect_args(username='uid', password='pwd',
                                           database='dbn' )
         keys = list(opts.keys())
@@ -399,6 +446,7 @@ class SQLAnyDialect(default.DefaultDialect):
             opts['host'] += ':%s' % opts['port']
             del opts['port']
         #
+        opts.update(dialect_opts)
         return ([], opts)
     #
 
@@ -414,14 +462,18 @@ class SQLAnyDialect(default.DefaultDialect):
 
     def _get_default_schema_name(self, connection):
         return connection.scalar(
-                     text("SELECT current user",
-                     typemap={'user_name': Unicode})
+                     text("SELECT current user").columns(column('user_name', Unicode))
              )
 
     def initialize(self, connection):
         super(SQLAnyDialect, self).initialize(connection)
         self.max_identifier_length = 128
 
+        VERSION_SQL = text('select @@version')
+        result = connection.execute(VERSION_SQL)
+        vers = result.scalar()
+        self.server_version_info = tuple( vers.split(' ')[0].split( '.' ) )
+
     def get_table_id(self, connection, table_name, schema=None, **kw):
         """Fetch the id for schema.table_name.
 
@@ -548,10 +600,12 @@ class SQLAnyDialect(default.DefaultDialect):
         column_cache[table_id] = columns
 
         REFCONSTRAINT_SQL = text("""
-          SELECT pt.table_name AS name, pt.table_id AS reftable_id
+          SELECT fk.foreign_index_id, i.index_name AS name, pt.table_id AS reftable_id
           FROM sys.sysfkey fk
           join sys.systab pt on fk.primary_table_id = pt.table_id
+          join sys.sysidx i on i.table_id=fk.primary_table_id
           WHERE fk.foreign_table_id = :table_id
+          and i.index_category=2
         """)
         referential_constraints = connection.execute(REFCONSTRAINT_SQL,
                                                      table_id=table_id)
@@ -564,6 +618,7 @@ class SQLAnyDialect(default.DefaultDialect):
 
         for r in referential_constraints:
             reftable_id = r["reftable_id"]
+            foreign_index_id = r["foreign_index_id"]
 
             if reftable_id not in table_cache:
                 c = connection.execute(REFTABLE_SQL, table_id=reftable_id)
@@ -594,10 +649,12 @@ class SQLAnyDialect(default.DefaultDialect):
             join sys.sysidxcol pic on (fk.primary_index_id=pic.index_id and fk.primary_table_id=pic.table_id)
             WHERE fk.primary_table_id = :reftable_id
             and fk.foreign_table_id = :table_id
+            and fk.foreign_index_id = :foreign_index_id
             """)
             ref_cols = connection.execute(REFCOLS_SQL,
                                           table_id=table_id,
-                                          reftable_id=reftable_id)
+                                          reftable_id=reftable_id,
+                                          foreign_index_id=foreign_index_id)
             for rc in ref_cols:
                 constrained_columns.append(columns[rc["fokey"]])
                 referred_columns.append(reftable_columns[rc["refkey"]])
@@ -666,6 +723,10 @@ class SQLAnyDialect(default.DefaultDialect):
         pks = results.fetchone()
         results.close()
 
+        if not pks:
+            return {"constrained_columns": [],
+                    "name": None}
+
         PKCOL_SQL = text("""
              select tc.column_name as col
              FROM sys.sysidxcol ic
@@ -786,3 +847,12 @@ class SQLAnyDialect(default.DefaultDialect):
             return False
         else:
             return True
+
+    def is_disconnect(self, e, connection, cursor):
+        """
+        Signal to SQLAlchemy whether *e* indicates that *connection* is
+        broken and the pool needs to be recycled.
+        """
+        if isinstance(e, sqlanydb.OperationalError):
+            return e.args[1] in (-101, -308,)
+        return False
